from jsonschema import draft4_format_checker, validate
from jsonschema.exceptions import ValidationError
from jsonschema.validators import Draft4Validator

from pyramid_swagger_spec.errors import APIError


class Output:
    def __init__(self, d):
        self.d = d

    def __json__(self, request):
        return self.d


class JSchema:
    def __init__(self, schema):
        self.schema = schema

    def get_json_schema(self):
        return self.schema

    def output(self, data):
        validator = Draft4Validator(
            schema=self.schema,
            format_checker=draft4_format_checker
        )
        try:
            validator.validate(data)
        except ValidationError as e:
            raise APIError(500, message="The server generated invalid output: %s" % (e.message,))
        else:
            return Output(data)



from collections import namedtuple
from json.decoder import JSONDecodeError

from jsonschema import draft4_format_checker, validate
from jsonschema.exceptions import ValidationError
from jsonschema.validators import Draft4Validator

from .errors import APIError
from .swagger import Types


def boolify(input):
    if input in (0, False, "False", "false", "0", None):
        return False
    else:
        return True


def _validate(spec, value):
    if spec["type"] == Types.number:
        value = int(value)
    if spec["type"] == Types.boolean:
        value = boolify(value)

    validate(schema={
        "type": spec["type"],
        "format": spec.get("format", None)
    }, instance=value)
    return value


ValidatedParams = namedtuple('ValidatedParams', ['path', 'query', 'body'])


def validate_request(request, api):
    path_values = dict()
    query_values = dict()
    body_values = dict()

    for x in api["parameters"]:
        if x["in"] == "body":
            validator = Draft4Validator(
                schema=x["schema"],
                format_checker=draft4_format_checker
            )
            try:
                validator.validate(request.json_body)
            except ValidationError as e:
                raise APIError(400, message="Body: %s" % (e.message,))
            except JSONDecodeError as e:
                raise APIError(400, message="Body: JSON could not be decoded.")
            else:
                body_values = request.json_body

        elif x["in"] == "query":
            # query param
            try:
                query_values[x["name"]] = _validate(x, request.GET[x["name"]])
            except ValidationError as e:
                raise APIError(400, message="Query: %s invalid. %s" % (x["name"], e.message))
            except KeyError:
                raise APIError(400, message="Query: %s missing" % (x["name"],))
        elif x["in"] == "path":
            try:
                path_values[x["name"]] = _validate(x, request.matchdict[x["name"]])
            except ValidationError as e:
                raise APIError(400, message="Path: %s invalid. %s" % (x["name"], e.message))
            except KeyError:
                raise APIError(400, message="Path: %s missing" % (x["name"],))
    return ValidatedParams(
        path=path_values,
        query=query_values,
        body=body_values
    )


# -*- coding: utf-8 -*-
from pyramid.view import view_config
from pyramid.response import Response
import json


class APIError(Exception):
    def __init__(self, code, status="err", message=""):
        self.code = code
        self.status = status
        self.message = message


def json_exception_view(exc, request):
    s = json.dumps({
        "status": exc.status,
        "message": exc.message,
    })
    response = Response(s)
    response.status_int = exc.code
    response.headers["Access-Control-Allow-Origin"] = "*"
    return response



from collections import defaultdict
from pyramid.path import DottedNameResolver
import functools
import inspect
import venusian
import zope.interface
from pyramid.renderers import render_to_response
from zope.interface.declarations import implementer

# copied from tomb_routes, modified

ACCEPT_RENDERER_MAP = {
    'json': 'application/json',
    'string': 'text/plain',
}


class MatchdictMapper(object):
    def __init__(self, **kwargs):
        self.view_settings = kwargs
        self.attr = self.view_settings.get('attr')
        self.blacklist = [
            'optional_slash',
        ]

    def __call__(self, view):
        @functools.wraps(view)
        def wrapper(context, request):
            kwargs = request.matchdict.copy()
            for k in self.blacklist:
                if k in kwargs:
                    del kwargs[k]

            if inspect.isclass(view):
                arg_len = len(inspect.getargspec(view.__init__).args)
                if arg_len == 2:
                    inst = view(request)
                elif arg_len == 3:
                    inst = view(context, request)
                else:
                    raise Exception("Class should accept `context` and "
                                    "`request` args only")
                meth = getattr(inst, self.attr)
                return meth(**kwargs)
            else:
                return view(request, **kwargs)

        return wrapper





def add_simple_route(
        config, path, target,
        append_matchdict=True,
        default_accept='text/html',
        route_kwargs={},
        *args, **kwargs
):
    """Configuration directive that can be used to register a simple route to
    a view.

    Examples:

    with view callable::

        config.add_simple_route(
            '/path/to/view', view_callable,
            renderer='json'
        )

    with dotted path to view callable::

        config.add_simple_route(
            '/path/to/view', 'dotted.path.to.view_callable',
            renderer='json'
        )
    """

    target = DottedNameResolver().maybe_resolve(target)
    mapper = config.get_routes_mapper()
    route_name = target.__name__
    route_name_count = 0

    if 'accept' in kwargs:
        val = kwargs.pop('accept')
        route_kwargs['accept'] = val
    else:
        # Disable */* by default, only accept 'text/html'
        renderer = kwargs.get('renderer', 'html')
        acceptor = ACCEPT_RENDERER_MAP.get(renderer, default_accept)
        route_kwargs['accept'] = acceptor

    if 'attr' in kwargs:
        route_name += '.' + kwargs['attr']

    routes = {route.name: route for route in mapper.get_routes()}
    orig_route_name = route_name

    while route_name in routes:
        route_name = '%s_%s' % (orig_route_name, route_name_count)
        route_name_count += 1

    current_pregen = kwargs.pop('pregenerator', None)

    orig_route_prefix = config.route_prefix
    # We are nested with a route_prefix but are trying to
    # register a default route, so clear the route prefix
    # and register the route there.
    if (path == '/' or path == '') and config.route_prefix:
        path = config.route_prefix
        config.route_prefix = ''

    config.add_route(
        route_name, path, pregenerator=current_pregen,
        **route_kwargs
    )

    kwargs['route_name'] = route_name

    if append_matchdict and 'mapper' not in kwargs:
        kwargs['mapper'] = MatchdictMapper

    config.add_view(target, *args, **kwargs)
    request_method = kwargs.get("request_method", "GET")
    if request_method.upper() != "OPTIONS":
        config.add_view(options_view, *args, **dict(kwargs, request_method="OPTIONS"))
    config.commit()
    config.route_prefix = orig_route_prefix


def options_view(request, *args, **kw):
    return request.response


def create_api_namespace(namespace):
    namespace = namespace.strip("/")

    class DRoute(object):
        def __init__(self, path, *args, **kwargs):
            """Constructor just here to accept parameters for decorator"""
            self.path = path
            view_name = kwargs.get("name", "")
            if len(view_name.strip("/")) == 0:
                raise Exception("You must provide a view_name to prevent route conflicts!")
            self.route_path = self.path.rstrip("/") + "/" + view_name.lstrip("/") if view_name else self.path
            self.prefixed_route_path = "/" + namespace + "/" + self.route_path.lstrip("/")
            self.args = args
            self.kwargs = kwargs

        def __call__(self, wrapped):
            """Attach the decorator with Venusian"""
            args = self.args
            kwargs = self.kwargs

            def callback(scanner, _name, wrapped):
                """Register a view; called on config.scan"""
                config = scanner.config

                # pylint: disable=W0142
                add_simple_route(config, self.prefixed_route_path, wrapped, *args, route_kwargs=dict(traverse=self.prefixed_route_path), **kwargs)

                request_method = kwargs.get("request_method", "GET")
                registry = config.registry
                registry.getUtility(IRouteRegistry).register(namespace, self.route_path, request_method, kwargs.get("api",{}))

            info = venusian.attach(wrapped, callback)

            if info.scope == 'class':  # pylint:disable=E1101
                # if the decorator was attached to a method in a class, or
                # otherwise executed at class scope, we need to set an
                # 'attr' into the settings if one isn't already in there
                if kwargs.get('attr') is None:
                    kwargs['attr'] = wrapped.__name__
            return wrapped

    return DRoute


class IRouteRegistry(zope.interface.Interface):
    registrations = zope.interface.Attribute("""blahblah""")

    def register(namespace, url, method, params):
        """bar blah blah"""


@implementer(IRouteRegistry)
class RouteRegistry:
    def __init__(self):
        self.registrations = defaultdict(lambda: defaultdict(dict))

    def register(self, namespace, url, method, params):
        self.registrations[namespace][url][method.lower()] = params



import time

from .errors import APIError, json_exception_view
from .namespace import RouteRegistry
from .validator import validate_request
from .schema import JSchema

def includeme(config):
    # Add a view-deriver to allow the "api" parameter to add_view
    # Also measure the performance and add CORS header
    def api_view_deriver(view, info):
        api = info.options.get('api')
        if api:
            def wrapper_view(context, request):
                if request.method.upper() != "OPTIONS":
                    start = time.time()
                    request.api = api
                    request.validated_params = validate_request(request, api)
                    response = view(context, request)
                    end = time.time()
                    response.headers['Access-Control-Allow-Origin'] = '*'
                    response.headers['X-View-Performance'] = '%.3f' % (end - start,)
                elif request.method.upper() == "OPTIONS":
                    response = view(context, request)
                    response.headers['Access-Control-Allow-Origin'] = '*'
                    response.headers['Access-Control-Allow-Headers'] = request.headers['Access-Control-Request-Headers']
                    response.headers['Access-Control-Allow-Methods'] = "DELETE, GET, HEAD, OPTIONS, PATCH, POST, PUT"
                    response.headers['Access-Control-Allow-Credentials'] = "true"
                return response

            return wrapper_view
        return view

    api_view_deriver.options = ('api',)
    config.add_view_deriver(api_view_deriver)

    # Register the RouteRegistry that will be used to store the api spec
    config.registry.registerUtility(RouteRegistry())

    config.add_view(view=json_exception_view, context=APIError)




from jsonschema.validators import Draft4Validator
from pyramid.response import Response
from pyramid.settings import asbool
from .namespace import IRouteRegistry


def generate_swagger(title, version, host, basePath, paths, schemes=["https",]):
    return {
        "swagger": "2.0",
        "host": host,
        "info": {
            "description": "",
            "version": version,
            "title": title,
            "contact": {},
            "license": {
                "name": "Proprietary"
            },
        },
        "schemes": schemes,
        "basePath": basePath,
        "paths": {
            itemUrl: {
                itemMethod: {
                    "tags": [methodParams.get("tag", "default"),],
                    "operationId": methodParams.get("operationId", ""),
                    "parameters": methodParams.get("parameters", []),
                    "consumes": methodParams.get("consumes", ["application/json", ]),
                    "produces": methodParams.get("produces", ["application/json", ]),
                    "summary": methodParams.get("summary", ""),
                    "description": methodParams.get("description", ""),
                    "responses": methodParams.get("responses", {}),
                } for itemMethod, methodParams in itemMethods.items()
            } for itemUrl, itemMethods in paths.items()
        }
    }


def create_swagger_view(config, namespace, title, version):
    namespace = namespace.strip("/")
    route_name = namespace+"_swagger"
    route_url = namespace + "/_swagger"
    config.add_route(route_name, route_url, request_method="GET")

    def view(request, *args, **kw):
        registry = request.registry
        route_registry = registry.getUtility(IRouteRegistry)
        registrations = route_registry.registrations[namespace]
        request.response.headers['Access-Control-Allow-Origin'] = '*'
        use_http = asbool(request.registry.settings.get("swagger.use_http_scheme", 0))
        host = request.registry.settings.get("swagger.host", "localhost:6543")
        schemes = []
        if use_http:
            schemes += ["http"]
        else:
            schemes += ["https"]
        return generate_swagger(title=title, version=version, host=host, basePath="/"+namespace, paths=registrations, schemes=schemes)

    config.add_view(view, route_name=route_name, request_method="GET", renderer="json")

    html_tmpl = """
<!DOCTYPE html>
<html>
  <head>
    <title>ReDoc</title>
    <!-- needed for adaptive design -->
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <meta name="robots" content="noindex" />

    <!--
    ReDoc doesn't change outer page styles
    -->
    <style>
      body {
        margin: 0;
        padding: 0;
      }
    </style>
  </head>
  <body>
    <redoc spec-url='%(url)s'></redoc>
    <script src="https://rebilly.github.io/ReDoc/releases/latest/redoc.min.js"> </script>
  </body>
</html>
    """

    html_route_name = namespace + "_swagger.html"
    html_route_url = namespace + "/_swagger.html"
    config.add_route(html_route_name, html_route_url, request_method="GET")

    def html_view(request, *args, **kw):
        return Response(body=html_tmpl % {
            "url": request.route_url(route_name)
        }, content_type="text/html")

    config.add_view(html_view, route_name=html_route_name, request_method="GET", renderer="json")


class Types:
    null = "null"
    boolean = "boolean"
    object = "object"
    array = "array"
    number = "number"
    string = "string"
    file = "file"


class Formats:
    hostname = "hostname"
    ipv4 = "ipv4"
    ipv6 = "ipv6"
    email = "email"
    uri = "uri"  # requires rfc3987
    datetime = "date-time"  # requires strict-rfc3339
    date = "date"
    time = "time"
    regex = "regex"
    color = "color"  # required webcolors
    byte = "byte"


class Patterns:
    uuid = "^[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}$"


def api(tag,
        operation_id,
        summary,
        parameters=[],
        consumes=["application/json"],
        produces=["application/json"],
        description="",
        responses={}):
    return {
        'tag': tag,
        'operationId': operation_id,
        'summary': summary,
        'parameters': parameters,
        'consumes': consumes,
        'produces': produces,
        'description': description,
        'responses': responses,
    }

def process_null_type(ret):
    if ret["x-nullable"]:
        if isinstance(ret["type"], list) and not Types.null in ret["type"]:
            ret["type"].append(Types.null)
        else:
            ret["type"] = [ret["type"], Types.null]
    return ret

def path_parameter(name, parameter_type, format="", description=""):
    return {
        "name": name,
        "in": "path",
        "required": True,
        "type": parameter_type,
        "description": description,
        "format": format
    }


def query_parameter(name, parameter_type, format="", required=False, description=""):
    return {
        "name": name,
        "in": "query",
        "required": required,
        "type": parameter_type,
        "description": description,
        "format": format
    }


def property(type, format="", nullable=False, description="", pattern=None):
    ret = {
        "type": type,
        "format": format,
        "x-nullable": nullable,
        "description": description,
    }

    if pattern:
        ret["pattern"] = pattern

    return process_null_type(ret)


def object_property(properties, nullable=False, description=""):
    return process_null_type({
        "type": "object",
        "x-nullable": nullable,
        "properties": properties,
        "description": description
    })


def array_property(items, nullable=False, description=""):
    return process_null_type({
        "type": "array",
        "x-nullable": nullable,
        "items": items,
        "description": description
    })


def body_parameter(schema, description=""):
    gschema = {
        "$schema": "http://json-schema.org/schema#",
        "type": "object",
        "properties": schema,
    }
    required = [k for k, v in schema.items() if not v.get("x-nullable")]
    if len(required) > 0:
        gschema["required"] = required

    Draft4Validator.check_schema(gschema)

    return {
        "name": "body",
        "in": "body",
        "required": True,
        "description": description,
        "schema": gschema
    }


def response(schema, description=""):
    schema = {
        "$schema": "http://json-schema.org/schema#",
        "type": "object",
        "properties": schema
    }

    Draft4Validator.check_schema(schema)

    return {
        "description": description,
        "schema": schema
    }


def file_response(description=""):
    schema = {
        "type": "file",
    }

    return {
        "description": description,
        "schema": schema
    }


